﻿namespace MyCompany.Crt.Managers
{
    using Microsoft.ApplicationInsights;
    using Microsoft.ApplicationInsights.DataContracts;
    using Microsoft.Dynamics.Commerce.Runtime;
    using Microsoft.Extensions.Caching.Memory;
    using MyCompany.Crt.Common;
    using System;
    using System.Diagnostics;
    using System.Globalization;
    using System.Threading.Tasks;

    /// <summary>
    /// A memory cache abstraction based on keys of types of string
    /// </summary>
    /// <typeparam name="TItem"></typeparam>
    public abstract class MemoryCacheAccessor<TItem>
    {
        private string _accessorUniqueName;

        // per stackoverflow, MemoryCache is threadsafe
        private static readonly Lazy<MemoryCache> lazy = new Lazy<MemoryCache>(() => new MemoryCache(new MemoryCacheOptions { }));

        private static MemoryCache cache
        {
            get
            {
                return lazy.Value;
            }
        }

        private MemoryCacheAccessor()
        {
        }

        protected MemoryCacheAccessor(string accessorUniqueName)
        {
            ThrowIf.NullOrWhiteSpace(accessorUniqueName, "accessorUniqueName");

            this._accessorUniqueName = accessorUniqueName;
        }

        private string MakeUniqueKey(object key)
        {
            return string.Format("{0}_{1}", _accessorUniqueName, key);
        }

        public bool TryGetValue(string key, TelemetryClient logger, out TItem value)
        {
            ThrowIf.NullOrWhiteSpace(key, "key");
            string key1 = MakeUniqueKey(key);

            var stopwatch = Stopwatch.StartNew();
            if (cache.TryGetValue<TItem>(key1, out value))
            {
                stopwatch.Stop();
                var trace = logger.CreateTraceWithSessionContext(SeverityLevel.Information);
                trace.Properties.Add("executionTimeMilliseconds", stopwatch.ElapsedMilliseconds.ToString(CultureInfo.InvariantCulture));
                trace.Properties.Add("MemoryCacheAccessor_CacheMode", "hit");
                trace.Properties.Add("EventId", "100000");
                trace.Message = string.Format("{0}.TryGetValue: Fetched item from cache for key '{1}'.", this._accessorUniqueName, key1);
                logger.TrackTrace(trace);
                return true;
            }
            else
            {
                stopwatch.Stop();
                var trace = logger.CreateTraceWithSessionContext(SeverityLevel.Information);
                trace.Properties.Add("executionTimeMilliseconds", stopwatch.ElapsedMilliseconds.ToString(CultureInfo.InvariantCulture));
                trace.Properties.Add("MemoryCacheAccessor_CacheMode", "miss");
                trace.Properties.Add("EventId", "100000");
                trace.Message = string.Format("{0}.TryGetValue: Cache item for key '{1}' was not found.", this._accessorUniqueName, key1);
                logger.TrackTrace(trace);
                return false;
            }
        }

        public TItem GetOrCreate(string key, Func<string, TItem> factory, TimeSpan? absoluteExpirationRelativeToNow, TimeSpan? slidingExpiration, TelemetryClient logger)
        {
            ThrowIf.NullOrWhiteSpace(key, "key");

            TItem value;
            if (this.TryGetValue(key, logger, out value))
            {
                return value;
            }
            else
            {
                var stopwatch = Stopwatch.StartNew();

                TItem newValue = factory(key);

                stopwatch.Stop();

                if (newValue != null)
                {
                    var trace = logger.CreateTraceWithSessionContext(SeverityLevel.Information);
                    trace.Properties.Add("executionTimeMilliseconds", stopwatch.ElapsedMilliseconds.ToString(CultureInfo.InvariantCulture));
                    trace.Properties.Add("EventId", "100000");
                    trace.Message = string.Format("{0}.GetOrCreate: Called factory method to fetch value for key '{1}'. Method returned a non-null value.", this._accessorUniqueName, MakeUniqueKey(key));
                    logger.TrackTrace(trace);

                    this.Set(key, newValue, absoluteExpirationRelativeToNow, slidingExpiration, logger);
                }
                else
                {
                    var trace = logger.CreateTraceWithSessionContext(SeverityLevel.Warning);
                    trace.Properties.Add("executionTimeMilliseconds", stopwatch.ElapsedMilliseconds.ToString(CultureInfo.InvariantCulture));
                    trace.Properties.Add("EventId", "100000");
                    trace.Message = string.Format("{0}.GetOrCreate: Called factory method to fetch value for key '{1}'. Method returned a null.", this._accessorUniqueName, MakeUniqueKey(key));
                    logger.TrackTrace(trace);
                }

                return newValue;
            }
        }

        private MemoryCacheEntryOptions MakeMemoryCacheOptions(TimeSpan? absoluteExpirationRelativeToNow, TimeSpan? slidingExpiration)
        {
            if (absoluteExpirationRelativeToNow.HasValue == slidingExpiration.HasValue)
            {
                throw new ArgumentNullException("TimeSpan? absoluteExpirationRelativeToNow, TimeSpan? slidingExpiration", "One of them has to be set");
            }

            var o = new MemoryCacheEntryOptions();
            if (absoluteExpirationRelativeToNow.HasValue)
            {
                o.AbsoluteExpirationRelativeToNow = absoluteExpirationRelativeToNow.Value;
            }
            else
            {
                o.SlidingExpiration = slidingExpiration;
            }

            return o;
        }

        public void Set(string key, TItem value, TimeSpan? absoluteExpirationRelativeToNow, TimeSpan? slidingExpiration, TelemetryClient logger)
        {
            ThrowIf.NullOrWhiteSpace(key, "key");

            if (value == null)
            {
                throw new ArgumentNullException("value");
            }

            string key1 = MakeUniqueKey(key);
            var stopwatch = Stopwatch.StartNew();
            cache.Set<TItem>(key1, value, options: MakeMemoryCacheOptions(absoluteExpirationRelativeToNow, slidingExpiration));
            stopwatch.Stop();
            var trace = logger.CreateTraceWithSessionContext(SeverityLevel.Information);
            trace.Properties.Add("executionTimeMilliseconds", stopwatch.ElapsedMilliseconds.ToString(CultureInfo.InvariantCulture));
            trace.Properties.Add("MemoryCacheAccessor_CacheMode", "set");
            trace.Properties.Add("EventId", "100000");
            if (absoluteExpirationRelativeToNow.HasValue)
            {
                trace.Properties.Add("MemoryCacheAccessor_AbsoluteExpirationInSeconds", absoluteExpirationRelativeToNow.Value.TotalSeconds.ToString(CultureInfo.InvariantCulture));
            }
            else
            {
                trace.Properties.Add("MemoryCacheAccessor_SlidingExpirationInSeconds", slidingExpiration.Value.TotalSeconds.ToString(CultureInfo.InvariantCulture));
            }
            trace.Message = string.Format("{0}.Set: Set new cached value for key '{1}'.", this._accessorUniqueName, key1);
            logger.TrackTrace(trace);
        }
    }
}
