namespace MyCompany.Crt.RequestHandlers
{
    using Microsoft.Dynamics.Commerce.Runtime;
    using Microsoft.Dynamics.Commerce.Runtime.Messages;
    using Microsoft.Dynamics.Commerce.Runtime.RealtimeServices.Messages;
    using MyCompany.Crt.Common;
    using MyCompany.Crt.Managers;
    using MyCompany.Crt.Messages;
    using System;
    using System.Threading.Tasks;

    public class SaveCartRequestHandler : SingleAsyncRequestHandler<SaveCartRequest>
    {
        protected override async Task<Response> Process(SaveCartRequest request)
        {
            ThrowIf.Null(request, "request");
            var telemetryLogger = Logger.GetLogger(request.RequestContext);

            // caching example 1: Response caching inside own handler
            // An overridden handler that takes care of caching for all responses of the original handler. In some cases this is an advantage over caching at the caller.
            // GetUserDefinedSecretStringValueRealtimeRequest may be a good example to cache in the handler and not at every caller. 
            var getUserDefinedSecretStringValueRealtimeRequest = new GetUserDefinedSecretStringValueRealtimeRequest("secret1");
            await request.RequestContext.Runtime.ExecuteAsync<Response>(getUserDefinedSecretStringValueRealtimeRequest, request.RequestContext, false);
            System.Diagnostics.Trace.TraceInformation("after getUserDefinedSecretStringValueRealtimeRequest");


            // caching example 2: Response caching outside of own handler
            // Caching is controlled at the caller. This way we could have different behaviour for different callers dependent on the caller. 
            var someResponseCacheAccessor = GetSomeResponseCacheAccessor();
            var expensiveHandlerResponse = someResponseCacheAccessor.GetOrCreate(request.Cart.Id, v =>
            {
                return DoSomethingExpensive(request).Result;
            }, absoluteExpirationRelativeToNow: null, slidingExpiration: TimeSpan.FromMinutes(20), telemetryLogger);

            // caching example 3: simple caching outside of own handler
            // Caching is controlled at the caller. A simple valuetype is being cached (in case no Response is available to cache)
            var externalIdCacheAccessor = GetExternalIdCacheAccessor();
            var expensiveCallReturn = externalIdCacheAccessor.GetOrCreate(request.Cart.Id, v =>
            {
                return GetExternalIdExpensive(request.Cart.Id).Result;
            }, absoluteExpirationRelativeToNow: TimeSpan.FromMinutes(20), slidingExpiration: null, telemetryLogger);

            // calling original next-in-line handler for SaveCartRequest
            return await request.RequestContext.Runtime.ExecuteNextAsync<Response>(this, request, request.RequestContext, false);
        }

        private async Task<Response> DoSomethingExpensive(SaveCartRequest request)
        {
            // pretending to call some other async handler and it takes 5 s to return
            return await Task.Run(() =>
            {
                System.Threading.Thread.Sleep(5000);
                return new SomeResponse();
            });
        }

        private async Task<int> GetExternalIdExpensive(string id)
        {
            // pretending to call some other code and it takes 5 s to return
            return await Task.Run(() =>
            {
                System.Threading.Thread.Sleep(5000);
                return 1000;
            });
        }

        private IResponseCacheAccessor GetSomeResponseCacheAccessor()
        {
            return new SomeResponseCacheAccessor();
        }

        private IExternalIdCacheAccessor GetExternalIdCacheAccessor()
        {
            return new ExternalIdCacheAccessor();
        }
    }
}